/* we'll just use core window system libraries, no bloat */
#include <X11/Xlib.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>
#include <stdint.h>
#include <assert.h>
#include <stdio.h>
#include <err.h>
#include "defs.h"
#include "gl.h"
/* do wywalenia */
#include <unistd.h>
#include <stdlib.h>

#define NAME "Versus' GPU Plugin"
#define AUTHOR "Johny Versus"
#define VERSION 1
#define REVISION 0
#define BUILD 0

/* dma off, ready for commands, GPU idle, display disabled */
/* interlace off, 15 bit, 256x240 */
#define INITIAL_STATUS 0x14000000

#define WIDTH 800
#define HEIGHT 600

/* current data available when reading the status register */
uint32_t gpu_status;
/* current data available when reading the data register */
static uint32_t gpu_data;
uint16_t vram[MAXX][MAXY];
/* VRAM transfer stuff */
/* current location during transfer */
uint16_t vtx, vty, vtxs;
/* coordinates of the lower bottom end of image to transfer */
uint16_t vtxe, vtye;
uint8_t image_transfer = 0;
int trcnt;
#define TRANSFER_BUFLEN 10


void
print_matrix(float *matrix)
{
	int j;

	for (j = 0; j < 16; j+= 4)
		printf("%8.4f %8.4f %8.4f %8.4f\n", matrix[j], matrix[j + 1],
			   matrix[j + 2], matrix[j + 3]);
}

void
print_matrices(void)
{
	float matrix[16];	
	int viewport[4];
	int i, n, once = 0;
	GLboolean enabled;

	glGetIntegerv(GL_VIEWPORT, viewport);
	printf("viewport: x=%d y=%d, w=%d h=%d\n", viewport[0],
		   viewport[1], viewport[2], viewport[3]);
	glGetFloatv(GL_PROJECTION_MATRIX, matrix);
	puts("current projection matrix:");
	print_matrix(matrix);
	puts("current modelview matrix:");
	glGetFloatv(GL_MODELVIEW_MATRIX, matrix);
	print_matrix(matrix);
	glGetIntegerv(GL_MAX_CLIP_PLANES, &n);
	n += GL_CLIP_PLANE0;
	for (i = GL_CLIP_PLANE0; i < n; i++) {
		glGetBooleanv(i, &enabled);
		if (enabled)
			printf("clip plane %d enabled\n", i - GL_CLIP_PLANE0), once = 1;
	}
	if (!once)
		puts("No clip planes enabled");
	glGetBooleanv(GL_TEXTURE_2D, &enabled);
	if (enabled)
		puts("2D texture mapping enabled");
	else
		puts("2D texture mapping disabled");
	glGetTexEnviv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, &i);
	printf("Current texture env mode: ");
	switch (i) {
	case GL_MODULATE:
		puts("GL_MODULATE");
		break;
	case GL_DECAL:
		puts("GL_DECAL");
		break;
	case GL_BLEND:
		puts("GL_BLEND");
		break;
	case GL_REPLACE:
		puts("GL_REPLACE");
		break;
	}
#if 0
	puts("current texture matrix:");
	glGetFloatv(GL_TEXTURE_MATRIX, matrix);
	print_matrix(matrix);
#endif
	printf("Current matrix mode is ");
	glGetIntegerv(GL_MATRIX_MODE, &i);
	switch (i) {
	case GL_MODELVIEW:
		puts("modelview.");
		break;
	case GL_PROJECTION:
		puts("projection.");
		break;
	case GL_TEXTURE:
		puts("texture.");
		break;
	case GL_COLOR:
		puts("color.");
		break;
	}
}

char *
PSEgetLibName(void)
{
	return NAME;
}

unsigned long
PSEgetLibType(void)
{
	return PSE_LT_GPU;
}

unsigned long
PSEgetLibVersion(void)
{
	return VERSION<<16 | REVISION<<8 | BUILD;
}

long GPUinit(void)
{
	/* read config file */

	return 0;
}

long GPUshutdown(void)
{

	return 0;
}

#define CHKW 64
#define CHKH 64
GLubyte check_image[CHKH][CHKW][4];

void
make_check_image(void)
{
    int i, j;
    GLubyte c;

    for (i = 0; i < CHKH; i++)
        for (j = 0; j < CHKW; j++) {
            c = ((((i & 8) == 0) ^ (((j & 8) == 0)))) * 255;
            check_image[i][j][0] = c;
            check_image[i][j][1] = c;
            check_image[i][j][2] = c;
            check_image[i][j][3] = c;
        }
}

/* XXX czesc zamienic na wartosci automatyczne */
static Display *dpy;
static XVisualInfo *vi;
static GLXContext cx;
static XSetWindowAttributes swa;
static Window win;

int32_t
GPUopen(Display  **display, char *caption, char *config)
{
	int gl_mode[] = {GLX_DOUBLEBUFFER, GLX_RGBA, GLX_DEPTH_SIZE, 24, None};
	/* create window, etc */
	Colormap cmap;
	void prepare_texture(int tx, int ty, int clx, int cly, int type);

	dpy = XOpenDisplay(NULL);
	if (dpy == NULL)
		return -1;
	if (!glXQueryExtension(dpy, NULL, NULL))
		return -1;
	vi = glXChooseVisual(dpy, DefaultScreen(dpy), gl_mode);
	if (vi == NULL)
		return -1;
	if (vi->class != TrueColor)
		return -1;
	cx = glXCreateContext(dpy, vi, None, GL_TRUE);
	if (cx == NULL)
		return -1;
	cmap = XCreateColormap(dpy, RootWindow(dpy, vi->screen), vi->visual,
						   AllocNone);
	swa.colormap = cmap;
	swa.border_pixel = 0;
	swa.event_mask =  KeyPressMask | KeyReleaseMask;
	win = XCreateWindow(dpy, RootWindow(dpy, vi->screen), 0, 0, WIDTH, HEIGHT,
						0, vi->depth, InputOutput, vi->visual,
						CWBorderPixel | CWColormap | CWEventMask, &swa);
	XSetStandardProperties(dpy, win, NAME, caption, None,
						   /* argv */NULL, /* argc */0, NULL);
	glXMakeCurrent(dpy, win, cx);
	XMapWindow(dpy, win);

    glClearColor(1, 0, 0, 1);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	gpu_status = INITIAL_STATUS;
	*display = dpy;
	print_matrices();

	return 0;
}

long
GPUclose()
{
	/* destroy window, free memory, cleanup */

	puts("GPU close called");
	gpu_status = 0;
	glXMakeCurrent(dpy, None, NULL);
	glXDestroyContext(dpy, cx);
	XDestroyWindow(dpy, win);
	XCloseDisplay(dpy);

	return 0;
}

void
GPUupdateLace(void)
{
	int glerr;
	int stop = 0;

	while ((glerr = glGetError()) != GL_NO_ERROR) {
		warnx("updateLace: glError detected: %s", gluErrorString(glerr));
		stop = 1;
	}
	if (stop)
		exit(1);
	glXSwapBuffers(dpy, win);
	/* glClear(GL_COLOR_BUFFER_BIT); */
}

int count = 0;
int check_status = 0;

unsigned long
GPUreadStatus(void) /* GP1 */
{
	count++;
	if (check_status) {
#ifdef DEBUG_STATUS
		printf("gpu.c: readStatus: called, check_status on, 0x%.8x\n",
			   gpu_status);
#endif
		check_status = 0;
	}

	return gpu_status;
}

void
GPUwriteStatus(unsigned long data) /* GP1 */
{
	uint8_t command = data >> 0x18 /* 24 */;
	int16_t x = 0, y = 0, x2, y2;
	int changed = 0;
	static int lastx, lasty;

	/* __asm__("int $3\n"); */
	check_status = 1;
#ifdef DEBUG_STATUS
	if (count)
		printf("gpu.c: read_status: read status called %d times.\n", count);
#endif
	count = 0;
	data &= 0xffffff;
#ifdef DEBUG_STATUS
	printf("gpu.c: writeStatus: command=0x%.2X, data=0x%.6lX\n",
		   command, data);
#endif
	/*printf("gpu.c: writeStatus: current status=0x%8lX, ", gpu_status);*/
	switch (command) {
	case 0:
		/* reset gpu */
		gpu_status = INITIAL_STATUS;
		changed = 1;
		break;
	case 1:
		/* reset the command buffer */
		break;
	case 2:
		/* reset the IRQ */
		break;
	case 3:
		/* display disable/enable */
		if (data)
			gpu_status &= ~(1<<0x17);
		else
			gpu_status |= (1<<0x17);
		break;
	case 4:
		/* dma setup */
		gpu_status &= 0x9fffffff;
		gpu_status |= data << 0x1d;
		if (data == 0 && image_transfer) {
			puts("killing image transfer due to DMA setup, recheck YOUR code");
			image_transfer = 0;
		}
		break;
	case 5:
		/* te trzy parametry to jakby overscan */
		/* sztucznie zwiekszaja ilosc wyswietlanych pikseli nie zwiekszajac */
		/* ich wielkosci */
		/* mozna by te wartosc uzyc do wywolania glOtho z niestandardowa */
		/* iloscia pikseli */
		/* start of display area */
		x = data & 0x3ff;
		y = (data >> 0xa) & 0x1ff;
#ifdef DEBUG_STATUS
		printf("gpu.c: writeStatus: cmd5: x=%d y=%d\n", x, y);
#endif
		break;
	case 6:
		/* horizontal display range */
		x = data & 0xfff;
		x2 = (data >> 0xc) & 0xfff;
#ifdef DEBUG_STATUS
		printf("gpu.c: writeStatus: cmd6: x=%d x2=%d\n", x, x2);
#endif
		x = x2 - x;
		break;
	case 7:
		/* vertical display range */
		y = data & 0x3ff;
		y2 = (data >> 0xa) & 0x3ff;
#ifdef DEBUG_STATUS
		printf("gpu.c: writeStatus: cmd7: y=%d y2=%d\n", y, y2);
#endif
		y = y2 - y;
		break;
	case 8:
		/* display mode */
		gpu_status &= 0xff00ffff;
		/* width0 and height, videomode, isrgb24 and isinter */
		gpu_status |= (data & 0x3f) << 0x11;
		/* width1 */
		gpu_status |= (data & 0x40) << (0x10 - 0x6);
		changed = 1;
		break;
	case 0x10:
		/* gpu info */
#ifdef DEBUG_STATUS
		printf("gpu_info cmd=0x%lX, ", data);
#endif
		switch (data) {
#if 0
		case 3:
			/* starting point of display */
			gpu_data = (dtlx << 0xc) | dtly;
			break;
		case 4:
			/* ending point of display */
			gpu_data = ((dtlx + disp_x) << 0xc) | (dtly + disp_y);
			break;
		case 5:
			/* drawing offset */
			gpu_data = disp_x << 0xa;
		case 6:
			gpu_data = disp_y << 0xa;
			break;
#endif
		case 7:
			/* gpu type */
			gpu_data = 2;
			break;
		}
		break;
	}
	if (changed) {
		switch (gpu_status & 0x70000) {
		case 0:
			x = 256;
			break;
		case 0x20000:
			x = 320;
			break;
		case 0x10000:
			x = 384;
			break;
		case 0x40000:
			x = 512;
			break;
		case 0x60000:
			x = 640;
			break;
		}
		y = (gpu_status & 80000) ? 480 : 240;
		if (x != lastx || y != lasty) {
			printf("gpu.c: writeStatus: switching virtual resolution to %d*%d\n", x, y);
			glMatrixMode(GL_PROJECTION);
			glLoadIdentity();
			glOrtho(0, x, y, 0, 0, 1);
			glMatrixMode(GL_MODELVIEW);
			lastx = x; lasty = y;
			print_matrices();
		}
	}
#ifdef DEBUG_STATUS
	printf("gpu.c: write_status: status changed to 0x%8X\n", gpu_status);
#endif
}

uint32_t
GPUreadData(void) /* GP0 */
{
	if (image_transfer == 2) {
		/* reading data from VRAM */
		gpu_data = *(uint32_t *)&vram[vtx][vty];
		vtx++;
		if (vtx == vtxe) {
			/* oops, even read, we have to correct */
#ifdef DEBUG
			printf("gpu.c: readData: even read: %dx%d\n", vtx, vty);
#endif
			gpu_data = vram[vtx - 1][vty] | (vram[0][vty + 1] << 0xa);
			vtx = vtxs, vty++;
		}
		/* printf("gpu.c: readData -> MEM(0x%.8X) (cnt=%d)\n", gpu_data, trcnt); */
		trcnt += 2;
		vtx++;
		if (vtx == vtxe)
			vtx = vtxs, vty++;
		if (vty == vtye) {
			/* not ready to receive image */
			gpu_status &= ~0x08000000;
			image_transfer = 0;
#ifdef DEBUG
			puts("image_transfer <- VRAM ended");
#endif
		}
	}
#ifdef DEBUG
	else
		printf("gpu.c: readData: spurious readData encountered, 0x%X returned\n",
			   gpu_data);
#endif

	return gpu_data;
}

void
GPUwriteData(input_t data) /* GP0 */
{
	/* are we doing transfer now? */
	static uint8_t func, i;
	static int8_t transfer;
	static input_t transfer_buf[TRANSFER_BUFLEN];


	check_status = 1;
	/* it seems that the data register return what it gets */
	gpu_data = data.dw;
	/* image transfer takes precedence over other commands */
	/* since it is an unfinished command */
	if (image_transfer == 4) {
		/* mono-color polyline completion */
		if (data.dw == 0x55555555) {
			image_transfer = 0;
			glEnd();
#ifdef DEBUG_POLYLINE
			puts("mono-color polyline ended");
#endif

			return;
		}
		glVertex2s(data.w[0], data.w[1]);
	}
	else if (image_transfer == 5) {
		/* multi-color polyline */
		if (data.dw == 0x55555555) {
			image_transfer = 0;
			glEnd();
#ifdef DEBUG_POLYLINE
			puts("multi-color polyline ended");
#endif

			return;
		}
		glColor3ub(data.b[0], data.b[1], data.b[2]);
		image_transfer = 6;
	}
	else if (image_transfer == 6) {
		assert(data.dw != 0x55555555);
		glVertex2s(data.w[0], data.w[1]);
		image_transfer = 5;
	}
	else if (image_transfer == 1) {
		/* writing data to VRAM */
		/* printf("gpu.c: writeData: -> VRAM(0x%.8x) (cnt=%d x=%d y=%d)\n", data.dw, trcnt, vtx, vty); */
#if 1
		assert(vtx < MAXX);
		assert(vty < MAXY);
#endif
		vram[vtx++][vty] = data.w[0];
		trcnt += 2;
		if (vtx == vtxe) {
			vtx = vtxs, vty++;
			if (vty == vtye) {
				image_transfer = 0;
#ifdef DEBUG
				puts("image_transfer ->VRAM ended on uneven byte");
#endif

				return;
			}
		}
		vram[vtx++][vty] = data.w[1];
		if (vtx == vtxe) {
			vtx = vtxs, vty++;
			if (vty == vtye) {
				image_transfer = 0;
#ifdef DEBUG
				puts("image_transfer ->VRAM ended");
#endif
			}
		}
	}
	else {
		assert(transfer >= 0);
		if (!transfer) {
			i = 0;
			func = data.b[3]/* & 0xf8 NIET */;
			if (!function_dispatcher[func]) {
				printf("gpu.c: writeData: UNIMPLEMENTED display function encountered: 0x%x\n", func);
				fflush(stdout);
				__asm__("int $3\n");
			}
			data.b[3] = 0;
			transfer = command_length[func];
			
			if (transfer < 0) {
				puts("multi_line unimplemented");
				fflush(stdout);
				__asm__("int $3\n");
			}
			
			assert(transfer);
			assert(i < TRANSFER_BUFLEN);
			transfer_buf[i++] = data;
			transfer--;
		}
		else {
#ifdef DEBUG_WRITE
			printf("gpu.c: writeData: additional %d arg: %.8X\n", i, data.dw);
#endif
			transfer_buf[i++] = data;
			transfer--;
		}
		if (!transfer) {
#ifdef DEBUG_WRITE
			printf("gpu.c: writeData: running function 0x%X, len=%d\n",
				   func, command_length[func]);
#endif
			function_dispatcher[func](transfer_buf);
		}
	}
}

void
GPUreadDataMem(uint32_t *data, int count)
{
	int i;

	for (i = 0; i < count; i++)
		data[i] = GPUreadData();
}

void
GPUwriteDataMem(input_t *data, int count)
{
	int i;

	for (i = 0; i < count; i++)
		GPUwriteData(data[i]);
}

long
GPUdmaChain(unsigned char *base_addr, unsigned long addr)
{
	while (addr != 0xffffff) {
		int count = base_addr[addr + 3];

#ifdef DEBUG_DMA
		printf("gpu.c: dmaChain: addr = 0x%lX, size=%d\n", addr, count);
#endif
		/* second word is the first command in a series */
		GPUwriteDataMem((input_t *)(base_addr + addr + 4), count);
		addr = *((uint32_t *)(base_addr + addr)) & 0xffffff;
	}
			
	return 0;
}

long
GPUconfigure(void)
{
	/* completely ignored, dialog bload too heavy */

	return 0;
}

long
GPUtest(void)
{

	return 0;
}
